from tabulate import tabulate
#! *Working with tabulate*
#! See more about Tabulate use at https://github.com/astanin/python-tabulate

table = [["spam",42],["eggs",451],["bacon",0]]
headers = ["item", "qty"]
tabulate_tab = tabulate(table, headers, tablefmt="fancy_grid") #%tab


# -*- coding: utf-8 -*-
#-----------------------------------------------------------------
# Copyright (C) 2020, the Loge development team
#
# This file is part of Loge
# Loge is distributed under the terms of GNU General Public License
# The full license can be found in 'license.txt'
# Loge development team can be found in 'development.txt'
#-----------------------------------------------------------------

'''
---------------------------------------------------------------------
Browser module.
Browser is a window where results of processing script are visible.
---------------------------------------------------------------------
'''

from PyQt5.QtWidgets import QTextBrowser

class Browser(QTextBrowser):

    def scroll_to_relposition(self, relposition):
        absmaximum = self.verticalScrollBar().maximum()
        absposition = relposition * absmaximum
        self.verticalScrollBar().setValue(absposition)


# -*- coding: utf-8 -*-
#-----------------------------------------------------------------
# Copyright (C) 2017-2020, the Loge development team
#
# This file is part of Loge
# Loge is distributed under the terms of GNU General Public License
# The full license can be found in 'license.txt'
# Loge development team can be found in 'development.txt'
#-----------------------------------------------------------------

import subprocess
import traceback
import os

from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication, QMessageBox, QDialog, QInputDialog
from PyQt5.QtPrintSupport import QPrintDialog
from PyQt5.QtGui import QCursor

from loge.core.Script import Script
from loge.core.Shell import Shell
from loge.core.core_utils import abspath, get_html_from_memo, get_html_from_code, get_html_from_markdownfile

class Core():
    
    def __init__(self):
        self.Script = Script()
        self.Shell = Shell()
        self.Gui = None
        #---
        self.watcher = None
        self.timer = None
        #---
        self.set_watcher()
        self.set_timer()
        self.Shell.assign_code(self.Script)
        #---
        self.scrolls_data = {}

    def set_watcher(self):
        self.watcher = QtCore.QFileSystemWatcher()
        self.watcher.fileChanged.connect(self.script_changed)
        self.Script.set_watcher(self.watcher)

    def set_timer(self):
        # -- Timer --
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.TimerAction)

    def watcher_clicked(self,checked):
        if self.Script.script_path :
            if self.Gui.main_window.actions['watch_script'].isChecked():
                self.Script.activate_watcher()
            else:
                self.Script.deactivate_watcher()
        else:
            QMessageBox.information(None, 'Info', 'Please create or open script first')
            self.Gui.main_window.actions['watch_script'].setChecked(False)

    def script_changed(self,file):
        self.Script.reloadcode()
        self.refresh()

    #------------------------ 

    def refresh(self, waitcursor = True, silenceErrorMassage = False):
        successruned = False        
        message = None
        error_trace = None
        if waitcursor:
            QApplication.setOverrideCursor(QCursor(QtCore.Qt.WaitCursor))
        try:
            self.Script.parse()
            self.Shell.run_parsed()
            self.Gui.browser_reload(self.Shell.report_html)
            successruned = True
        except Exception as e:
            message = 'Some problem - ' + str(e)
            error_trace = str(traceback.format_exc())
        finally:
            QApplication.restoreOverrideCursor()
            self.Gui.warning_browser.hide()
            self.Gui.browser.setStyleSheet("background-color: White;")
        if not successruned and message:
            if silenceErrorMassage:
                self.Gui.warning_browser.show()
                self.Gui.warning_browser.setTextColor(QtCore.Qt.red)
                self.Gui.warning_browser.setText(message + '\n' + error_trace)
                self.Gui.browser.setStyleSheet("background-color: rgb(255, 240, 240);")
            else:
                QMessageBox.information(None, message, error_trace)
        #---
        apptitle=''
        if self.Script.script_path:
            apptitle = ' - ' + os.path.basename(self.Script.script_path)
        if self.Script.saveLock:
            apptitle += '  (read only)'
        apptitle = self.Gui.get_app_main_title() + apptitle
        self.Gui.set_app_title(apptitle)
        #---
        return successruned
    
    def synchronizeEditorToScript(self):
        if self.Script.code_oryginal != self.Gui.editor.toPlainText():
            self.Gui.editor.setPlainText(self.Script.code_oryginal)

    def synchronizeScriptToEditorAndRefresh(self):
        self.Script.code_oryginal = self.Gui.editor.toPlainText() 
        self.refresh(silenceErrorMassage=True)

    def reload_script_file(self):
        if self.Script.script_path :
            self.Script.reloadcode()
            self.synchronizeEditorToScript()
            self.refresh()
        else:
            QMessageBox.information(None, 'Info', 'Please create or open script first')

    #------------------------ 

    def file_new(self):
        template_path = abspath('templates/x_newtemplate.py')
        self.Shell.delete_tmpfile() #Cleaning Loge tmp files
        self.file_save_if_changed()#save current file if changed
        self.Script.newFile(template_path, 'Save new script as', 'newScript.py')
        self.synchronizeEditorToScript()
        self.refresh()
        #--File browser update
        self.Gui.file_browser.reload()
            
    def file_open(self, file_path = None, saveLock = False, browser_path_update=True):
        self.Shell.delete_tmpfile() #cleaning Loge tmp files
        self.scrolls_archive_positions()#archive scroll position
        self.file_save_if_changed()#save current file if changed
        self.Script.openFile(file_path, saveLock)
        self.synchronizeEditorToScript()
        self.refresh()
        self.scrols_load_position_from_archive()#load scroll position
        #--File browser update
        if browser_path_update:
            self.Gui.file_browser.reload()
        #--Timer
        self.TimerFromCode()

    def file_openreadonly(self, file_path = None):
        self.file_open(file_path = None, saveLock = True)

    def file_save(self):
        if self.Script.script_path :
            self.Script.code_oryginal = self.Gui.editor.toPlainText()
            if self.Script.savecode():
                self.Gui.main_window.showStatusInfo('Saved to '+ self.Script.script_path)
        else:
            self.file_saveas()

    def file_saveas(self):
        self.scrolls_archive_positions()#archive scroll position
        self.Script.saveAs()
        self.refresh()
        #--File browser update
        self.Gui.file_browser.reload()
    
    def file_save_if_changed(self):
        if self.Script.code_has_changed():
            # - ask if save changes
            dialogReply = QMessageBox.question( None, 'Loge', "Do you want save changes?", 
                                                QMessageBox.Yes | QMessageBox.No, 
                                                QMessageBox.No)
            if dialogReply == QMessageBox.Yes:
                self.file_save()

    def scrolls_archive_positions(self):
        if self.Script.script_path:
            scroll_editor = self.Gui.editor.verticalScrollBar().value()
            scroll_browser = self.Gui.browser.verticalScrollBar().value()
            #---
            self.scrolls_data[self.Script.script_path] = [scroll_editor, scroll_browser]

    def scrols_load_position_from_archive(self):
        if self.Script.script_path in self.scrolls_data:
            scroll_editor = self.scrolls_data[self.Script.script_path][0]
            self.Gui.editor.verticalScrollBar().setValue(scroll_editor)
            #---
            scroll_browser = self.scrolls_data[self.Script.script_path][1]
            self.Gui.browser.verticalScrollBar().setValue(scroll_browser)
        else:
            self.Gui.editor.verticalScrollBar().setValue(0)
            self.Gui.browser.verticalScrollBar().setValue(0)
            
    def scrolls_synchronize(self):
        if self.Gui.main_window.actions['editor_scroll_synchro'].isChecked():
            self.Gui.browser.scroll_to_relposition(self.Gui.editor.get_scroll_relposition())
        
    def show_file_browser(self):
        if self.Gui.main_window.actions['file_browser'].isChecked():
            self.Gui.file_browser.show()
            self.Gui.file_browser.reload()
        else:
            self.Gui.file_browser.hide()

    #------------------------ 

    def file_edit(self):
        if self.Script.saveLock:
            QMessageBox.information(None, 'Info', 'File open read only. Use save as option')
            return False
        if self.Gui.main_window.actions['file_edit'].isChecked():
            self.Gui.main_window.splitter_editor.show()
            self.Gui.toolbar_editor.setVisible(True)
            self.synchronizeEditorToScript()
        else:
            self.Gui.main_window.splitter_editor.hide()
            self.Gui.toolbar_editor.setVisible(False)

    #------------------------ 

    def show_python_source(self):
        html_page = get_html_from_code(self.Script.code_oryginal)
        self.Gui.set_browser_content(html_page)

    def show_html(self):
        html_page = get_html_from_code(self.Shell.report_html)
        self.Gui.set_browser_content(html_page)

    def show_markdown(self):
        html_page = get_html_from_code(self.Shell.report_markdown)
        self.Gui.set_browser_content(html_page)

    def show_loge(self):
        self.Gui.set_browser_content(self.Shell.report_html)

    def show_syntax(self):
        html_page = get_html_from_memo('x_syntax.md')
        self.Gui.set_browser_content(html_page)

    #------------------------
    def PreviewMarkdown(self):
        #---asking for file path
        filename = self.Script.open_markdown()
        if not filename == '':
            #---
            html_page = get_html_from_markdownfile(filename)
            self.Gui.set_browser_content(html_page)

    def SaveMarkdown(self):
        if self.Script.script_path :
            initname = os.path.basename(self.Script.script_path).replace('.py', '.md')
            self.Shell.save_report_markdown(self.Script.savedir, initname)
        else:
            QMessageBox.information(None, 'Info', 'Please create or open script first')

    #------------------------ 
            
    def startpage(self):
        html_page = get_html_from_memo('x_startpage.md')
        self.Gui.set_browser_content(html_page)

    def help(self):
        html_page = get_html_from_memo('x_help.md')
        self.Gui.set_browser_content(html_page)

    def about(self):
        html_page = get_html_from_memo('x_about.md')
        self.Gui.set_browser_content(html_page)

    #------------------------ 

    def file_print(self):
        dialog = QPrintDialog()
        if dialog.exec_() == QDialog.Accepted:
            self.Gui.get_browser_document().print_(dialog.printer())
    
    #------------------------ 

    def tutorial(self):
        tutorial_path = abspath('memos/x_tutorial.py')
        savedir = self.Script.savedir #coping current savedir
        #--opening oryginal tutorial
        self.Shell.delete_tmpfile()
        self.Script.openFile(tutorial_path)
        #---saving temporary tutorial copy
        tutorial_tmppath = os.path.join(self.Shell.tmpdir, 'tmp_tutorial.py')
        self.Script.saveAs(tutorial_tmppath)
        #---opening temporary tutorial copy
        self.Shell.delete_tmpfile()
        self.Script.openFile(tutorial_tmppath)
        self.Script.savedir = savedir #back to previouse savedir
        self.refresh()
        self.file_edit()

    #------------------------ 

    def floatprecision(self):
        if self.Script.script_path :
            #---asking for precision as int number
            value = QInputDialog.getInt(    None, 
                                            'Float display precysion', 'Set the precison:',
                                            value = self.Shell.float_display_precison,
                                            min = 1, max = 9, step = 1)[0]
            #---
            self.Shell.float_display_precison = value
            self.refresh()
        else:
            QMessageBox.information(None, 'Info', 'Please create or open script first')

    #-----------------------

    def TimerButtonClicked(self):
        if self.Script.script_path:
            if self.Gui.toolbar_timer.timerButton.isChecked():
                self.TimerStart()
            if not self.Gui.toolbar_timer.timerButton.isChecked():
                self.TimerStop()
        else:
            self.TimerStop()
            QMessageBox.information(None, 'Info', 'Please create or open script first')

    def TimerAction(self):
        if self.Script.script_path:
            if self.Gui.toolbar_timer.timerButton.isChecked():
                self.Gui.toolbar_timer.timerPulse.setChecked(not(self.Gui.toolbar_timer.timerPulse.isChecked()))
                self.refresh(False)

    #---------------------
    def TimerStart(self):
        timespace = int(self.Gui.toolbar_timer.timerSpinBox.value() * 1000)
        self.timer.start(timespace)
        self.Gui.toolbar_timer.timerButton.setChecked(True)
        self.Gui.toolbar_timer.timerSpinBox.setEnabled(False)
        self.Gui.toolbar_timer.timerPulse.setEnabled(True)
        self.refresh(False)
                
    def TimerStop(self):
        self.timer.stop()
        self.Gui.toolbar_timer.timerButton.setChecked(False)
        self.Gui.toolbar_timer.timerSpinBox.setEnabled(True)
        self.Gui.toolbar_timer.timerPulse.setChecked(False)
        self.Gui.toolbar_timer.timerPulse.setEnabled(False)

    def TimerFromCode(self):
        parameters = self.Script.getTimerParameters()
        if parameters:
            self.Gui.toolbar_timer.timerButton.setChecked(parameters[0])
            self.Gui.toolbar_timer.timerSpinBox.setValue(parameters[1])
            if parameters[0]:
                self.TimerStart()
        else:
            self.TimerStop()
            self.Gui.toolbar_timer.timerButton.setChecked(False)
            self.Gui.toolbar_timer.timerSpinBox.setValue(2.0)



import matplotlib.pyplot as figure
import numpy as np

#! ####*Quadratic function y = ax^2 + bx + c*

#%img parabola.png

#! Quadratic function coefficients

possible_a_list = [-10.0, -2.0, 5.0, 20.0 ]
a = possible_a_list[3] #<< - set a coefficient

b = 9 #<< - set b coefficient
c = -40 #<< - set c coefficient

# The roots (zeros) calculateing
delta = b**2 - 4*a*c #%requ
if delta >= 0:
    #! ###There are roots :) 
    x1 = (-b + delta**0.5) / (2 * a) #%requ - first root
    x2 = (-b - delta**0.5) / (2 * a) #%requ - second root
else:
    #! ###Roots not exist :(
    x1 = 'not exist'
    x2 = 'not exist'

# Our quadratic function plot
xi_from = -3 #<< - plot form x value
xi_to = 4 #<< - plot to x value
xi = np.arange(xi_from, xi_to, 1)
yi = a * xi**2 + b * xi + c
figure.grid()
figure.plot(xi, yi) #%plt
figure.clf()



# -*- coding: utf-8 -*-
#-----------------------------------------------------------------
# Copyright (C) 2017-2020, the Loge development team
#
# This file is part of Loge
# Loge is distributed under the terms of GNU General Public License
# The full license can be found in 'license.txt'
# Loge development team can be found in 'development.txt'
#-----------------------------------------------------------------

import os

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSlot

from loge.core.core_utils import APP_PATH


class Tree(QtWidgets.QTreeView):
	def __init__(self, parent=None):
		super(Tree, self).__init__(parent)
		#---
		self.setMaximumWidth(200)
		self.clicked.connect(self.on_treeView_clicked)
		#---
		self.core = None

	def assing_core (self, core):
		self.core = core
		
	def reload(self):
		#---Dir to show
		if self.core.Script.script_path:
			dir_to_show = os.path.dirname(self.core.Script.script_path)
		else:
			dir_to_show=APP_PATH
		#---Link the tree to a model
		model = QtWidgets.QFileSystemModel()
		model.setRootPath(dir_to_show)
		model.setNameFilters( ["*.py"] )
		self.setModel(model)
		#---Set the tree's index to the root of the model
		indexRoot = model.index(model.rootPath())
		self.setRootIndex(indexRoot)
		#---Hide tree size and date columns
		self.hideColumn(1)
		self.hideColumn(2)
		self.hideColumn(3)
		#---Hide tree header
		self.setHeaderHidden(True)

	@pyqtSlot(QtCore.QModelIndex)
	def on_treeView_clicked(self, index):
		indexItem = self.model().index(index.row(), 0, index.parent())
		#---
		fileName = str(self.model().fileName(indexItem))
		filePath = str(self.model().filePath(indexItem))
		#---
		if os.path.isfile(filePath):
			if os.path.splitext(filePath)[1] == '.py':
				self.core.file_open(filePath, browser_path_update=False)



# -*- coding: utf-8 -*-
#-----------------------------------------------------------------
# Copyright (C) 2017-2020, the Loge development team
#
# This file is part of Loge
# Loge is distributed under the terms of GNU General Public License
# The full license can be found in 'license.txt'
# Loge development team can be found in 'development.txt'
#-----------------------------------------------------------------

"""
---------------------------------------------------------------------
The module containing a main application window class.
---------------------------------------------------------------------
"""
import time

from PyQt5.QtWidgets import QDesktopWidget, QMainWindow, QStyle, QSplitter
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtCore import Qt, QSize

from loge.gui.gui_utils import get_icon, create_action
from loge.gui.gui_utils import translate, TR_GUI_CONTEXT

# a scale of application window width to cover the % of the whole screen available space
_SW_WIDTH = 0.5
# a scale of application window height to cover the % of the whole screen available space
_SW_HEIGHT = 0.7
# application window minimum width in pixels
_MIN_W_WIDTH = 100
# application window minimum height in pixels
_MIN_W_HEIGHT = 100
# application logo icon
_APP_ICON = "logo.png"


class MainWindow(QMainWindow):

    def __init__(self,app_name,app_version):
        super().__init__()
        self.actions = {} #actions dictionary besides QWidget.actions() list
        self._init_main_window()
        self.main_title = app_name + ' (' + app_version+')'
        self.setWindowTitle(self.main_title)
        self._set_actions()
        # ---
        self.splitter = QSplitter(Qt.Horizontal)
        self.setCentralWidget(self.splitter)

        self.splitter_editor = QSplitter(Qt.Vertical)
        #---
        self.status = self.statusBar()


    def _init_main_window(self):
        # -- Window settings --
        ag = QDesktopWidget().availableGeometry()
        window_size = ag.size()
        app_width = int(_SW_WIDTH * window_size.width())
        app_height = int(_SW_HEIGHT * window_size.height())
        window_size.scale(app_width,app_height,Qt.IgnoreAspectRatio)
        self.setGeometry(
        QStyle.alignedRect(Qt.LeftToRight,
            Qt.AlignCenter,
            window_size,
            ag
        ))
        self.setMinimumSize(_MIN_W_WIDTH,_MIN_W_HEIGHT)
        self.setWindowIcon(get_icon(_APP_ICON))

    def _set_actions(self):
        self.actions['file_new'] = create_action('new.png','New',QKeySequence.New
                            ,'Create new *.py script from template',self)

        self.actions['file_open'] = create_action('open.png','Open',QKeySequence.Open
                              ,'Open existing *.py script',self)

        self.actions['file_openreadonly'] = create_action('openreadonly.png','Open read only'
                                      ,None,'Open existing *.py script - edit and saving not aloved',self)

        self.actions['file_save'] = create_action('save.png','Save',QKeySequence.Save
                               ,'Save changes to script file',self)

        self.actions['file_saveas'] = create_action('saveas.png','Save as',QKeySequence.SaveAs
                                 ,'Save current script',self)

        self.actions['file_browser'] = create_action('file_browser.png','File browser',QKeySequence.SaveAs
                                 ,'It opens/closes file browser',self)
        self.actions['file_browser'].setCheckable(True)

        self.actions['reload_script_file'] = create_action('reload.png','Reload',QKeySequence.Refresh
                                 ,'Reload *.py script',self)

        self.actions['watch_script'] = create_action('watch_file.png','Reload',QKeySequence.Refresh
                                 ,'If pushed report will be reloaded if script file has changed',self)
        self.actions['watch_script'].setCheckable(True)

        self.actions['file_edit'] = create_action('edit.png','Edit', None
                              ,'It opens/closes editor for current script',self)
        self.actions['file_edit'].setCheckable(True)

        self.actions['show_html'] = create_action('showhtml.png','Show report as HTML',None
                                  ,'It shows report as HTML code',self)

        self.actions['show_markdown'] = create_action('showmarkdown.png','Show report as Markdown',None
                                     ,'It shows report as Markdown code',self)

        self.actions['show_source'] = create_action('showsource.png','Show python source',None
                                    ,'It shows python source code',self)

        self.actions['show_loge'] = create_action('showreport.png','Back to Loge report',None
                                   ,'Back to Loge report',self)

        self.actions['preview_markdown'] = create_action('previewmarkdown.png','Preview some Markdown',None
                                   ,'You can preview some Markdown document - it not change watched *.py script',self)

        self.actions['save_markdown'] = create_action('savemarkdown.png','Save report as Markdown file',None
                                   ,'Save raport as Markdown file',self)

        self.actions['print'] = create_action('print.png',translate(TR_GUI_CONTEXT,'Print document'),QKeySequence.Print
                               ,translate(TR_GUI_CONTEXT,'Print document'),self)

        self.actions['help'] = create_action('help.png',translate(TR_GUI_CONTEXT,'Help'),QKeySequence.HelpContents
                              ,translate(TR_GUI_CONTEXT,'Help information'),self)

        self.actions['about'] = create_action('about.png',translate(TR_GUI_CONTEXT,'About')
                               ,None,translate(TR_GUI_CONTEXT,'Loge project information'),self)

        self.actions['floatprecision'] = create_action(None,translate(TR_GUI_CONTEXT,'Float precision'),None
                                        ,translate(TR_GUI_CONTEXT,'Set float display precision'),self)

        self.actions['syntax'] = create_action('syntax.png',translate(TR_GUI_CONTEXT,'Syntax help'),'F2'
                                ,translate(TR_GUI_CONTEXT,'Show Loge syntax help'),self)

        self.actions['tutorial'] = create_action('tutorial.png',translate(TR_GUI_CONTEXT,'Tutorial'),None
                                  ,translate(TR_GUI_CONTEXT,'Open tutorial script'),self)

        self.actions['editor_zoom_in'] = create_action('zoom_in.png','Edytor zoom in',None
                                   ,'Edytor zoom in',self)

        self.actions['editor_zoom_out'] = create_action('zoom_out.png','Edytor zoom out',None
                                   ,'Edytor zoom out',self)

        self.actions['editor_refresh_report'] = create_action('refresh.png','Refresh report','Ctrl+E'
                                   ,'Refresh report [Ctrl+E]',self)

        self.actions['editor_auto_refresh_report'] = create_action('refresh_auto.png','Auto refresh report',None
                                   ,'Auto refresh report',self)
        self.actions['editor_auto_refresh_report'].setCheckable(True)

        self.actions['editor_scroll_synchro'] = create_action('scroll.png','Scroll synchro',None
                                   ,'If pushed, it synchronize editor and report scrolls',self)
        self.actions['editor_scroll_synchro'].setCheckable(True)

        self.actions['editor_insert_greek_letter'] = create_action('greek.png','Insert Greek letter','Ctrl+G'
                                   ,'It replace the letter before cursor into Greek letter, for example a into α. [Ctrl+G]',self)

        self.actions['editor_insert_unicode_prime_character'] = create_action('prime.png','Insert prime character',"Ctrl+'"
                                   ,"It insert unicode version of prime character that can be used in variable name [Ctrl+']",self)

    def set_action_slot(self,action_name,slot):
        """
        Sets a slot to an action.

        Args:
            action_name (str): the name of an action
            slot: a function that is called in response to a signal
        """
        self.actions[action_name].triggered.connect(slot)

    def showStatusInfo(self, massage=''):
        current_time_string = time.strftime("%H:%M:%S", time.gmtime())
        massage = 'at %s - %s'%(current_time_string, massage)
        self.status.showMessage(massage)



# -*- coding: utf-8 -*-
#-----------------------------------------------------------------
# Copyright (C) 2020, the Loge development team
#
# This file is part of Loge
# Loge is distributed under the terms of GNU General Public License
# The full license can be found in 'license.txt'
# Loge development team can be found in 'development.txt'
#-----------------------------------------------------------------

from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *

import loge.gui.syntax as syntax
from  loge.gui.greek_dict import greek_dict
 
class LineNumberArea(QWidget):
 
    def __init__(self, editor):
        super().__init__(editor)
        self.myeditor = editor
 
    def sizeHint(self):
        return Qsize(self.editor.lineNumberAreaWidth(), 0)
 
    def paintEvent(self, event):
        self.myeditor.lineNumberAreaPaintEvent(event)

class CodeEditor(QPlainTextEdit):
    def __init__(self, parent = None):
        super(CodeEditor,self).__init__(parent)
        self.highlight = syntax.PythonHighlighter(self.document())
        
        self.lineNumberArea = LineNumberArea(self)
 
        self.blockCountChanged.connect(self.updateLineNumberAreaWidth)
        self.updateRequest.connect(self.updateLineNumberArea)
        self.cursorPositionChanged.connect(self.highlightCurrentLine)
        
        self.updateLineNumberAreaWidth(0)
        
        self.setTabStopWidth(self.fontMetrics().width(' ') * 4)
        self.setStyleSheet("""QPlainTextEdit {background-color: white;
                            font-family: Courier New;}""")
        self.zoomIn(2)   
 
    def lineNumberAreaWidth(self):
        digits = 1
        count = max(1, self.blockCount())
        while count >= 10:
            count /= 10
            digits += 1
        space = 3 + self.fontMetrics().width('9') * digits
        return space
 
 
    def updateLineNumberAreaWidth(self, _):
        self.setViewportMargins(self.lineNumberAreaWidth(), 0, 0, 0)
 
 
    def updateLineNumberArea(self, rect, dy):
 
        if dy:
            self.lineNumberArea.scroll(0, dy)
        else:
            self.lineNumberArea.update(0, rect.y(), self.lineNumberArea.width(),
                       rect.height())
 
        if rect.contains(self.viewport().rect()):
            self.updateLineNumberAreaWidth(0)
 
 
    def resizeEvent(self, event):
        super().resizeEvent(event)
 
        cr = self.contentsRect();
        self.lineNumberArea.setGeometry(QRect(cr.left(), cr.top(),
                    self.lineNumberAreaWidth(), cr.height()))
        self.setTabStopWidth(self.fontMetrics().width(' ') * 4)
 
 
    def lineNumberAreaPaintEvent(self, event):
        mypainter = QPainter(self.lineNumberArea)
 
        mypainter.fillRect(event.rect(), Qt.lightGray)
 
        block = self.firstVisibleBlock()
        blockNumber = block.blockNumber()
        top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
        bottom = top + self.blockBoundingRect(block).height()
 
        # Just to make sure I use the right font
        height = self.fontMetrics().height()
        while block.isValid() and (top <= event.rect().bottom()):
            if block.isVisible() and (bottom >= event.rect().top()):
                number = str(blockNumber + 1)
                mypainter.setPen(Qt.black)
                mypainter.drawText(0, int(top), self.lineNumberArea.width(), height,
                 Qt.AlignRight, number)
 
            block = block.next()
            top = bottom
            bottom = top + self.blockBoundingRect(block).height()
            blockNumber += 1
 
    def highlightCurrentLine(self):
        extraSelections = []
 
        if not self.isReadOnly():
            selection = QTextEdit.ExtraSelection()
 
            lineColor = QColor(Qt.lightGray).lighter(120)
 
            selection.format.setBackground(lineColor)
            selection.format.setProperty(QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            extraSelections.append(selection)
        self.setExtraSelections(extraSelections)

    def get_scroll_relposition(self):
        absposition = self.verticalScrollBar().value()
        absmaximum = self.verticalScrollBar().maximum()
        if absmaximum != 0:
            relposition = 1.0 * absposition / absmaximum
        else:
            relposition = 0.0
        return relposition
    
    def insert_greek_letter(self):
        #what is the letter before cursor
        cursor_possition = self.textCursor().position()
        letter_before_cursor = self.toPlainText()[cursor_possition-1]
        if letter_before_cursor in greek_dict.keys():
            #what is the greek letter assigned to letter before cursor
            greek_letter = greek_dict[letter_before_cursor]
            #replace letter before cursor by the assigned greek letter
            self.textCursor().deletePreviousChar()
            self.insertPlainText(greek_letter)
        else:
            pass

    def insert_unicode_prime_character(self):
        self.insertPlainText("ʹ") # this is U+02B9 unicode prime character that can be used for variable name